import ctypes
import logging
from pathlib import Path

from pydantic import ByteSize

from .config.database import DatabaseConfig
from .config.engine import EngineConfig, EngineField, PUAConfig
from .config.scan import (
    GeneralConfig,
    HeuristicConfig,
    MailConfig,
    ParseConfig,
    ScanConfig,
)


class Client:
    def __init__(
        self,
        *,
        libclamav_path: str = "/opt/lib/libclamav.so",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        if not Path(libclamav_path).exists():
            raise FileNotFoundError(f"ClamAV library not found: {libclamav_path}")

        self.engine_config = EngineConfig()
        self.scan_config = ScanConfig()
        self.database_config = DatabaseConfig()
        self.logger = logger

        try:
            clamav = ctypes.CDLL(libclamav_path)
            if (ret := clamav.cl_init(ctypes.c_int(0x0))) != 0:
                self.logger.exception(
                    f"Failed to initialize ClamAV: {self.client.cl_strerror(ret)}"
                )
                raise RuntimeError("Failed to initialize ClamAV")
            engine = clamav.cl_engine_new()
            if not engine:
                self.logger.exception(
                    f"Failed to initialize ClamAV: {self.client.cl_strerror(ret)}"
                )
                raise RuntimeError("Failed to initialize ClamAV")
            self.engine = engine
            self.client = clamav
        except Exception as e:
            raise e

    def __del__(self):
        if hasattr(self, "client"):
            self.client.cl_engine_free(self.engine)

    def set_pua_conf(self, config: PUAConfig):
        if config.enabled:
            try:
                pua_categories = config.to_string()
                self.client.cl_engine_set_str(
                    self.engine,
                    EngineField["pua_categories"],
                    str(pua_categories).encode(),
                )
            except Exception as e:
                raise e

    def set_engine_conf(self, config: EngineConfig):
        try:
            for key, value in config.model_dump().items():
                key = EngineField[key]
                if isinstance(value, int) or isinstance(value, ByteSize):
                    self.client.cl_engine_set_num(
                        self.engine, key, ctypes.c_ulonglong(value)
                    )
                elif isinstance(value, str):
                    self.client.cl_engine_set_str(
                        self.engine, key, ctypes.c_char_p(value.encode())
                    )
                else:
                    raise ValueError(f"Invalid value type: {type(value)}")
        except Exception as e:
            self.logger.exception(f"Failed to set engine configuration: {config}")
            raise e

    def get_engine_conf(self) -> EngineConfig:
        config = {}
        for key, value in self.engine_config.model_dump().items():
            err = ctypes.c_uint()
            if isinstance(value, int) or isinstance(value, ByteSize):
                result = self.client.cl_engine_get_num(
                    self.engine, EngineField[key], ctypes.byref(err)
                )
                self.logger.debug(f"{key}: {result}")
            elif isinstance(value, str):
                self.client.cl_engine_get_str.restype = ctypes.c_char_p
                result = self.client.cl_engine_get_str(
                    self.engine, EngineField[key], ctypes.byref(err)
                )
            if err.value not in [0, 3]:
                raise RuntimeError(f"Failed to get engine configuration: {key}")
            config[key] = result
        return EngineConfig.model_validate(config)

    def set_database_conf(self, config: DatabaseConfig):
        self.database_config = config

    def set_scan_conf(self, config: ScanConfig):
        self.scan_config = config

    @classmethod
    def from_clamd_conf(
        cls,
        *,
        clamd_conf_path: str,
        libclamav_path: str = "/opt/lib/libclamav.so",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        try:
            config_dict = {}
            pua_config = PUAConfig()
            with Path(clamd_conf_path).open() as f:
                for line in f:
                    line = line.strip()
                    if not line or line.startswith("#"):
                        continue
                    key, value = line.split(None, 1)

                    if key == "ExcludePUA":
                        pua_config.excludes.append(value)
                    elif key == "IncludePUA":
                        pua_config.includes.append(value)

                    if value in ("yes", "true", "1"):
                        value = True
                    elif value in ("no", "false", "0"):
                        value = False

                    if key == "DetectPUA" and isinstance(value, bool):
                        pua_config.enabled = value

                    config_dict[key] = value
            engine_config = EngineConfig.model_validate(config_dict)
            database_config = DatabaseConfig.model_validate(config_dict)
            scan_config = ScanConfig(
                general=GeneralConfig.model_validate(config_dict),
                parse=ParseConfig.model_validate(config_dict),
                heuristic=HeuristicConfig.model_validate(config_dict),
                mail=MailConfig.model_validate(config_dict),
            )
            client = Client(
                libclamav_path=libclamav_path,
                logger=logger,
            )
            client.set_engine_conf(engine_config)
            client.set_pua_conf(pua_config)
            client.set_database_conf(database_config)
            client.set_scan_conf(scan_config)
            return client
        except FileNotFoundError as e:
            logger.exception(f"Clamd configuration file not found: {clamd_conf_path}")
            raise Exception from e

    def load_db(self, db_path: str = "/tmp/clamav"):
        if not Path(db_path).exists():
            raise FileNotFoundError(f"ClamAV database not found in: {db_path}")

        sigs = ctypes.c_uint32(0)
        if (
            ret := self.client.cl_load(
                db_path.encode("utf-8"),
                self.engine,
                ctypes.byref(sigs),
                self.database_config._to_bit_flag(),
            )
        ) != 0:
            self.logger.exception(
                f"Failed to load ClamAV database: {self.client.cl_strerror(ret)}"
            )
            self.client.cl_engine_free(self.engine)
            raise RuntimeError("Failed to load ClamAV database")
        if sigs.value == 0:
            self.logger.exception(f"No signatures loaded from {db_path}")
            raise RuntimeError("No signatures loaded")
        self.logger.info(f"Loaded {sigs.value} signatures")

    def compile_engine(self):
        if (ret := self.client.cl_engine_compile(self.engine)) != 0:
            self.logger.exception(
                f"Failed to load ClamAV database: {self.client.cl_strerror(ret)}"
            )
            self.client.cl_engine_free(self.engine)
            raise RuntimeError("Failed to compile ClamAV engine")
        self.logger.info("Compiled ClamAV engine. Ready to scan files")

    def scan_file(self, file_path: str) -> str | None:
        if not Path(file_path).exists():
            raise FileNotFoundError(f"File not found: {file_path}")
        _cl_scan_options = self.scan_config._to_c_struct()
        virname = ctypes.c_char_p()
        scanned = ctypes.c_ulong()
        result = self.client.cl_scanfile(
            file_path.encode("utf-8"),
            ctypes.byref(virname),
            ctypes.byref(scanned),
            self.engine,
            ctypes.byref(_cl_scan_options),
        )
        if result == 0:
            self.logger.info("No virus detected")
            return None
        elif result == 1 and virname.value:
            self.logger.info(f"Virus detected: {virname.value}")
            return virname.value.decode()
        else:
            raise RuntimeError(f"Failed to scan file: {result}")
